unit dyJPEG;

// adaptation ralise par tourlourou-Yves Lemaire-2008
// d'aprs LoadJPEG.cpp de shenron666 sur CodeS-SourceS
// (lui-mme d'aprs une source non prcisment nomme)
// et les exemples (LibJpeg) de Thomas G. Lane de l'IJG
// (Independent JPEG Group), ainsi que les explications
// de Cristi Cuturicu (tutoriel jpgalgo-code jpegdec.c)
// et les tutoriels de Calvin Hass sur ImpulseAdventure
// et son utilitaire JPEGsnoop de dcodage des fichiers

// code sans aucune garantie et librement rutilisable, 
// en mentionnant cet en-tte, et signalant les erreurs

// l'objet TdyYCbCrBaselineJPEGDecoder dcode les JPEG,
// cods par la Baseline Sequential DCT Compression, et
// uniquement dans l'espace de couleurs YCbCr (ou YUV),
// avec une prcision de 8 bits par composante, l'image
// tant passe dans un BitMap,  dtruire aprs usage.

// Unit dyJPEG, version 1.0 destine  tre acclre !

interface

uses
  Classes, SysUtils, Graphics, Windows, Math;

const
  // 3 octets par pixel
  BytesPerPixel=3;

  // Color Channels Index
  ccY=1; ccU=2; ccV=3;

type
  // donnes d'une composante de l'image = canal de couleur
  // le code propos se limite au dcodage des jpeg utilisant
  // le codage  3 composantes : une de luminance (Y)
  // et deux de chrominance (Cb=U et Cr=V)
  TColorChannel=record
    Id: Byte;                           // 1=Y 2=Cb=U 3=Cr=V (4=I 5=Q pour le codage YIQ)
    HorizontalSamplingFactor,           // nombre de DU (Data Unit) par MCU (Minimum Coded Unit) dans les
    VerticalSamplingFactor: Byte;       // 2 directions ; la taille de la MCU dpend du canal le plus prcis
    QT_Table, DC_Table, AC_Table: Byte; // indices des tables utilises pour le dcodage
    DcValue: SmallInt;                  // stockage de la valeur Dc, car dcodage de ses variations
  end;

  // la limitation  l'espace de couleur YUV permet d'utiliser un type ddi
  TYUV=record
   ChannelNb: Byte;  // indice de la composante (1=Y 2=Cb=U 3=Cr=V)
   HRes, VRes: Byte; // rsolutions relatives en nb de DU par MCU
   DUperMCU: Byte;   // Data Unit per Minimum Coded Unit
  end;

  // taille en pixels de la MCU (Minimum Coded Unit)
  TMCU=record
    Width,
    Height: Byte;
  end;

  // tables de Huffman, qui permettent le dcodage des coefficients DC et AC
  // le code ne reprsente pas la valeur directement, mais sa catgorie
  // cad le nombre de bits qui la composent,  lire  la suite
  PHuffmanTable=^THuffmanTable;
  THuffmanTable=record
    Bits: array[0.. 15] of Byte; // nombre de codes de taille i bits
    Hval: array[0..255] of Byte; // valeurs stockes dans l'ordre de l'arbre  reconstituer
    Size: array[0..255] of Byte; // taille du ime code
    Code: array[0..255] of Word; // code lui-mme
  end;

  // donnes non utilises dans la suite du code
  TProgressiveScan=record
    ss, se: Byte; // Spectral Selection (plage de coefficients DCT : ss=start=0 et se=end=63 pour les JPEG squentiels)
    ah, al: Byte; // Successive Approximation (High et Low bits pour le DCT ; 0 pour les JPEG squentiels)
  end;

  // type interne reprsentant un coefficient AC dcod
  // correspondant au nombre de zros prcdant la catgorie (taille)
  // de la valeur elle-mme, selon le principe du Zero Run Length Coding
  TAcTerm=record
  case Boolean of
    True : (EOB: Word);   // End Of Block si nul (<-> 0 prcd de 0 zros)
    False: (Zeros,        // nombre de zros prcdant la valeur
            Categ: Byte); // catgorie (nombre de bits) de la valeur
  end;

  // permet de fixer la prcision des rels (Extended inutiles)
  TMyFloat=Single; 

  // tableau de rels qui doit subir l'IDCT
  // (Inverse Discrete Cosine Transformation)
  // cad frquences spatiales -> valeurs elles-mmes
  TDctArray=array[0..63] of TMyFloat;

  // exceptions prives, gres ou non !
  EdyJpegError=class(Exception);

type
  // dcodeur de fichiers JPEG cods selon la
  // Baseline Sequential DCT JPG Compression
  // et dans l'espace de couleurs YCbCr=YUV
  TdyYCbCrBaselineJPEGDecoder=class
  private
    FFromFile: Boolean;                               // donnes
    FStream: TStream;                                 // lies  la
    FByteRead, FNextBitToRead: Byte;                  // lecture du fichier
    FSamplePrecision, FComponentsNumber: Byte;        // limitation  une prcision (FSamplePrecision) de 8 bits/canal
    FHeight, FWidth: Word;                            // dimensions de l'image
    FColorChannels: array[ccY..ccV] of TColorChannel; // canaux couleur dcods
    FYUV: array[ccY..ccV] of TYUV;                    // donnes des canaux couleur utiles au dcodage
    FMCU: TMCU;                                       // donnes relatives  la taille d'une MCU
    FRstInterval: Word;                               // nombre de MCU avant un marqueur Restart RSTn
    FQTable: array[0..2,0..63] of Word;               // tables de quantization        
    FAcTable, FDcTable: array[0..2] of THuffmanTable; // tables de Huffman des diffrentes composantes
    FProgressiveScan: TProgressiveScan;               // donnes non exploites
    FBuffer: Pointer;                                 // espace de dcompression
    FTicks: Cardinal;                                 // pour chronomtrage
    FBitmap: Graphics.TBitMap;                        // BitMap de sortie
  protected
    function GetBitmap: Graphics.TBitMap;
    function ReadByte: Byte;
    function ReadWord: Word;
    procedure ShlNextBitToWord(var AWord: Word);
    function ReadBitsToWord(Nb: Byte): Word;
    function BitsToInt(ABits: Byte; AValue: Word): SmallInt;
    function GetDcTermCategory(ADcTable: THuffmanTable): Byte;
    function GetAcTerm(AAcTable: THuffmanTable): TAcTerm;
    procedure AAN_IDCT(var ADct: TDctArray);
    procedure ReadMarkers;
  public
    constructor Create; overload;
    destructor Destroy; override;
    procedure Clear;
    procedure DecodeFromFile(AFile: string);
    procedure DecodeFromStream(AStream: TStream);
    property Ticks: Cardinal read FTicks;          // dure du dcodage (passage au BitMap inclus, lecture fichier exclue)
    property bmp: Graphics.TBitMap read GetBitmap; // BitMap que l'on peut rcuprer dans un TGraphic, par exemple
      //**************************************************************************************************
      // attention : celui qui accde  la proprit bmp devient responsable de la libration du BitMap \\
      //**************************************************************************************************
  end;

implementation

//uses Unit1; // uniquement pour dbuggage

const
  // masques de lecture des 16 bits d'un Word
  BitMask: array[1..16] of Word
           = (1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,16184,32768);

  // facteurs lis aux rotations de la transformation IDCT
  // valent 1.0 puis cos(k*Pi/16)*sqrt(2) pour k de 1  7
  AAN_Scale: array[0..7] of TMyFloat
           = (1.0, 1.387039845, 1.306562965, 1.175875602,
              1.0, 0.785694958, 0.541196100, 0.275899379);

  // parcours en zig-zag des indices des Data Units
  // qui permet d'augmenter les frquences spatiales
  // en loignant les pixels voisins
  ZigZag: array[0..63] of Byte
           = ( 0, 1, 5, 6,14,15,27,28,
               2, 4, 7,13,16,26,29,42,
               3, 8,12,17,25,30,41,43,
               9,11,18,24,31,40,44,53,
              10,19,23,32,39,45,52,54,
              20,22,33,38,46,51,55,60,
              21,34,37,47,50,56,59,61,
              35,36,48,49,57,58,62,63 );

type
  // pour conversions faciles
  TIntToByte=record
  case Boolean of
    True : (Int: integer);
    False: (Byte,b1,b2,b3: Byte);
  end;

  // pixel RGB ou YUV
  TYUV3=packed record
  case Boolean of
    True : (Triple: TRGBTriple);
    False: (Y, U, V: Byte);
  end;

  // ligne de pixels
  TPRT=array of TYUV3;

  // indices des registres de calcul et constantes pour l'algorithme AAN de l'IDCT
  // dclar ici pour l'initialisation des constantes
  AAN_index=(t00,t01,t02,t03,t04,t05,t06,t07,t10,t11,t12,t13,z05,z10,z11,z12,z13,ct1,ct2,ct3,ct4);

var
  AAN: array[t00..ct4] of TMyFloat; // tableau pour les calculs d'IDCT selon AAN
  AAN_PreCalc: TDctArray;           // facteurs lis  AAN et pralables  l'IDCT
  i,j,k: Byte;                      // pour boucles d'initialisation du tableau

S: string;


constructor TdyYCbCrBaselineJPEGDecoder.Create;
begin
  inherited;
  FFromFile:=False;
  FBuffer:=nil;
  FBitmap:=nil;
  Clear;
end;

destructor TdyYCbCrBaselineJPEGDecoder.Destroy;
begin
  Clear;
  inherited;
end;

procedure TdyYCbCrBaselineJPEGDecoder.Clear;
begin
  if FFromFile
  then begin
    FreeAndNil(FStream);    // cr en interne =>  librer
    FFromFile:=False;
  end
  else FStream:=nil;        // il avait t pass => pas  effacer
  if Assigned(FBuffer)
  then FreeMem(FBuffer);
  if Assigned(FBitmap)      // mis  nil par le Getter
  then FreeAndNil(FBitmap); // donc  dtruire si pas demand
  FRstInterval:=0;          // car pas forcment de section DRI pour le fixer
end;

function TdyYCbCrBaselineJPEGDecoder.GetBitmap: Graphics.TBitMap;
begin
  Result:=FBitmap; // renvoie le BitMap  celui qui l'a demand
  FBitmap:=nil;    // et le rend responsable de sa destruction
end;

function TdyYCbCrBaselineJPEGDecoder.ReadByte: Byte;
begin
  if FStream.Read(FByteRead,1)=1
  then begin
    Result:=FByteRead; // Bits 87654321
    FNextBitToRead:=0; // cad sur octet suivant, par dfaut
  end
  else raise EdyJpegError.Create('Stream ReadByte Error @'+IntToStr(FStream.Position));
end;

function TdyYCbCrBaselineJPEGDecoder.ReadWord: Word;
var
  w: Word;
begin
  if FStream.Read(w,2)=2
  then Result:=Swap(w)    // alignement Motorola (Big Endian) des Jpeg
  else raise EdyJpegError.Create('Stream ReadWord Error');
end;

procedure TdyYCbCrBaselineJPEGDecoder.ShlNextBitToWord(var AWord: Word);
begin
  if FNextBitToRead=0 // cad prochain octet ( ByteRead : 87654321 ; octet suivant : 0... )
  then begin
    ReadByte;
    if FByteRead=$FF // il faut tenir compte des exceptions
    then begin
      repeat ReadByte until FByteRead<>$FF;       // on saute les rptitions de $FF
      if FByteRead=0 then FByteRead:=$FF;        // $FF est cod par la succession $FF00
      if FByteRead in [$D0..$D7] then ReadByte; // Restart Markers : changent le cadre de lecture
    end;
    FNextBitToRead:=8; // donc nouvel octet
  end;
  AWord:=AWord shl 1;
  if FByteRead and BitMask[FNextBitToRead] > 0 then Inc(AWord);
  Dec(FNextBitToRead);
end;

function TdyYCbCrBaselineJPEGDecoder.ReadBitsToWord(Nb: Byte): Word;
var
  b: Byte;
begin
  Result:=0;
  for b:=1 to Nb do ShlNextBitToWord(Result);
end;

// la taille (catgorie) du coefficient DC est code selon Huffman
// qui exploite le Variable Length Coding : codage des valeurs
// les plus frquentes par les codes de plus petite taille 
function TdyYCbCrBaselineJPEGDecoder.GetDcTermCategory(ADcTable: THuffmanTable): Byte;
var
  i,j,index: Byte;
  w: Word;
begin
  w:=0;
  index:=0;
  i:=0;
  while i<16
  do begin
    ShlNextBitToWord(w);              // on lit le code bit par bit
    for j:=1 to ADcTable.Bits[i]      // puis on cherche dans ceux de sa longueur
    do begin
      if w=ADcTable.Code[index]       // s'il y est bien
      then begin
        Result:=ADcTable.Hval[index]; // auquel cas on renvoie la valeur qu'il code
        Exit;                         // qui donne la longueur en bits du coefficient
      end;
      Inc(index);
    end;
    Inc(i);
  end;
end;

// les coefficients AC sont cods selon le Zero Run Length Coding
// sur un octet : nb de 0 le prcdant, puis taille du coefficient
function TdyYCbCrBaselineJPEGDecoder.GetAcTerm(AAcTable: THuffmanTable): TAcTerm;
var
  i,j,index: Byte;
  w: Word;
begin
  w:=0;
  index:=0;
  i:=0;
  while i<16
  do begin
    ShlNextBitToWord(w);          // on lit le code bit par bit
    for j:=1 to AAcTable.Bits[i]  // puis on cherche dans ceux de sa longueur
    do begin
      if w=AAcTable.Code[index]   // s'il y est bien
      then begin
        i:=AAcTable.Hval[index];  // auquel cas on lit la valeur qu'il reprsente
        Result.Zeros:=i shr 4;    // chiffre hexa de poids fort : 0  15 zros prcdant
        Result.Categ:=i and $0F;  // chiffre hexa de poids faible : catgorie (taille) du coefficient cod
        Exit;                     // et on retourne le tout
      end;
      Inc(index);
    end;
    Inc(i);
  end;
end;

// traduction d'un code de taille donne en entier
function TdyYCbCrBaselineJPEGDecoder.BitsToInt(ABits: Byte; AValue: Word): SmallInt;
begin
  if AValue and BitMask[ABits] > 0                     // quelle est la valeur du 1 bit ?
  then Result:=AValue                                  // s'il commence par 1, il est positif...
  else Result := -( AValue xor (BitMask[ABits+1]-1) ); // sinon, il est ngatif => transfo complique !
end;

// AA&N algorithm of Inverse Discrete Cosine Transform
// Arai, Agui and Nakajima's algorithm for scaled IDCT
// il en existe des implmentations sur entiers, et il
// y a d'autres algorithmes, tels que celui selon LL&M
// le choix des rels est plus prcis, mais plus lent.
procedure TdyYCbCrBaselineJPEGDecoder.AAN_IDCT(var ADct: TDctArray);
var
  i: Byte;
begin
  // traitement par colonnes
  for i:=0 to 7
  do begin
    // lignes paires
    AAN[t00]:=ADct[i+0*8];
    AAN[t01]:=ADct[i+2*8];
    AAN[t02]:=ADct[i+4*8];
    AAN[t03]:=ADct[i+6*8];
    AAN[t10]:=AAN[t00]+AAN[t02];
    AAN[t11]:=AAN[t00]-AAN[t02];
    AAN[t13]:=AAN[t01]+AAN[t03];
    AAN[t12]:=AAN[ct1]*(AAN[t01]-AAN[t03])-AAN[t13];
    AAN[t00]:=AAN[t10]+AAN[t13];
    AAN[t03]:=AAN[t10]-AAN[t13];
    AAN[t01]:=AAN[t11]+AAN[t12];
    AAN[t02]:=AAN[t11]-AAN[t12];
    // lignes impaires
    AAN[t04]:=ADct[i+1*8];
    AAN[t05]:=ADct[i+3*8];
    AAN[t06]:=ADct[i+5*8];
    AAN[t07]:=ADct[i+7*8];
    AAN[z13]:=AAN[t06]+AAN[t05];
    AAN[z10]:=AAN[t06]-AAN[t05];
    AAN[z11]:=AAN[t04]+AAN[t07];
    AAN[z12]:=AAN[t04]-AAN[t07];
    AAN[t07]:=AAN[z11]+AAN[z13];
    AAN[t11]:=AAN[ct1]*(AAN[z11]-AAN[z13]);
    AAN[z05]:=AAN[ct2]*(AAN[z10]+AAN[z12]);
    AAN[t10]:=AAN[ct3]*AAN[z12]-AAN[z05];
    AAN[t12]:=AAN[ct4]*AAN[z10]+AAN[z05];
    AAN[t06]:=AAN[t12]-AAN[t07];
    AAN[t05]:=AAN[t11]-AAN[t06];
    AAN[t04]:=AAN[t10]+AAN[t05];
    // r-affectations
    ADct[i+0*8]:=AAN[t00]+AAN[t07];
    ADct[i+7*8]:=AAN[t00]-AAN[t07];
    ADct[i+1*8]:=AAN[t01]+AAN[t06];
    ADct[i+6*8]:=AAN[t01]-AAN[t06];
    ADct[i+2*8]:=AAN[t02]+AAN[t05];
    ADct[i+5*8]:=AAN[t02]-AAN[t05];
    ADct[i+4*8]:=AAN[t03]+AAN[t04];
    ADct[i+3*8]:=AAN[t03]-AAN[t04];
  end;
  // traitement par lignes
  for i:=0 to 7
  do begin
    // colonnes paires
    AAN[t00]:=ADct[0+i*8];
    AAN[t01]:=ADct[2+i*8];
    AAN[t02]:=ADct[4+i*8];
    AAN[t03]:=ADct[6+i*8];
    AAN[t10]:=AAN[t00]+AAN[t02];
    AAN[t11]:=AAN[t00]-AAN[t02];
    AAN[t13]:=AAN[t01]+AAN[t03];
    AAN[t12]:=AAN[ct1]*(AAN[t01]+AAN[t03])-AAN[t13];
    AAN[t00]:=AAN[t10]+AAN[t13];
    AAN[t03]:=AAN[t10]-AAN[t13];
    AAN[t01]:=AAN[t11]+AAN[t12];
    AAN[t02]:=AAN[t11]-AAN[t12];
    // colonnes impaires
    AAN[t04]:=ADct[1+i*8];
    AAN[t05]:=ADct[3+i*8];
    AAN[t06]:=ADct[5+i*8];
    AAN[t07]:=ADct[7+i*8];
    AAN[z13]:=AAN[t06]+AAN[t05];
    AAN[z10]:=AAN[t06]-AAN[t05];
    AAN[z11]:=AAN[t04]+AAN[t07];
    AAN[z12]:=AAN[t04]-AAN[t07];
    AAN[t07]:=AAN[z11]+AAN[z13];
    AAN[t11]:=AAN[ct1]*(AAN[z11]-AAN[z13]);
    AAN[z05]:=AAN[ct2]*(AAN[z10]+AAN[z12]);
    AAN[t10]:=AAN[ct3]*AAN[z12]-AAN[z05];
    AAN[t12]:=AAN[ct4]*AAN[z10]+AAN[z05];
    AAN[t06]:=AAN[t12]-AAN[t07];
    AAN[t05]:=AAN[t11]-AAN[t06];
    AAN[t04]:=AAN[t10]+AAN[t05];
    // r-affectations
    ADct[0+i*8]:=AAN[t00]+AAN[t07];
    ADct[7+i*8]:=AAN[t00]-AAN[t07];
    ADct[1+i*8]:=AAN[t01]+AAN[t06];
    ADct[6+i*8]:=AAN[t01]-AAN[t06];
    ADct[2+i*8]:=AAN[t02]+AAN[t05];
    ADct[5+i*8]:=AAN[t02]-AAN[t05];
    ADct[4+i*8]:=AAN[t03]+AAN[t04];
    ADct[3+i*8]:=AAN[t03]-AAN[t04];
  end;
end;

procedure TdyYCbCrBaselineJPEGDecoder.DecodeFromFile(AFile: string);
var
  tms: TMemoryStream;
begin
  Clear;
  tms:=TMemoryStream.Create;
  try
    tms.LoadFromFile(AFile);
    FFromFile:=True;
  except
    on E: Exception
    do raise EdyJpegError.Create(E.Message);
  end;
  DecodeFromStream(tms);
end;

// lecture des marqueurs pour identifier les sections
// cad dcodage des sections initialisant les donnes
// ainsi que les outils et tables ncessaires
procedure TdyYCbCrBaselineJPEGDecoder.ReadMarkers;
var
  Terminated: Boolean;
  marker, o, j, k: Byte;
  S: string;
  SectionLengthLeft: Word;
  HTable: PHuffmanTable;
  w: Word;
// dcodage de la section SOF0 => Baseline JPEG = donnes des canaux de couleurs
procedure Decode_SOF0;
  var
    b: Byte;
  begin
    FSamplePrecision:=ReadByte;  // nombre de bits par canal de "couleur"
    if FSamplePrecision<>8       // 12 galement valide dans la norme JPEG
    then raise EdyJpegError.Create('Color Channel Sample Precision Not Supported');
    FHeight:=ReadWord;           // dimensions
    FWidth:=ReadWord;            // de l'image
    FComponentsNumber:=ReadByte; // nombre de canaux <=> "couleurs" :
    if FComponentsNumber<>3      // 1=Grey Scaled 3=Color (YCbCr ou YIQ) 4=Color CMYK
    then raise EdyJpegError.Create('Only 3 Components Jpeg Supported Yet');
    if (SectionLengthLeft-6)<>(3*FComponentsNumber)
    then raise EdyJpegError.Create('Components Length Error');
    for b:=ccY to ccV
    do with FColorChannels[b] do begin
      Id:=ReadByte;              // nature du canal de "couleur" : 1=Y 2=Cb=U 3=Cr=V (4=I 5=Q)
      if not (Id in [ccY..ccV]) then raise EdyJpegError.Create('Color Channel Not Supported');
      if Id<>b then raise EdyJpegError.Create('Color Channels Order Error'); // => toujours YUV
      o:=ReadByte;               // lecture octet divis en 2 valeurs hexa (poids fort et faible)
      HorizontalSamplingFactor:=o shr 4; // taille relative du canal, => inverse
      VerticalSamplingFactor:=o and $0F; // de la prcision de l'chantillonnage
      QT_Table:=ReadByte;        // numro de la table de quantisation
      with FYUV[Id], FMCU
      do begin                   // initialisation du canal
        ChannelNb:=b;
        HRes:=HorizontalSamplingFactor;
        VRes:=VerticalSamplingFactor;
        DUperMCU:=HRes*VRes;     // nombre d'units par bloc               // utilit ????????????????????
        Width:=Max(Width,HRes);  // on cherche la dimension maximale
        Height:=Max(Height,VRes);// exprime en units (carrs de 8*8)
      end;
    end;
    FMCU.Width:=8*FMCU.Width;    // on convertit cette
    FMCU.Height:=8*FMCU.Height;  // dimension en pixels
  end;
// dcodage d'une section DHT = chargement d'une ou plusieurs tables d'Huffman
procedure Decode_DHT;
  var
    b,i: Byte;
  begin
    while SectionLengthLeft>0
    do begin
      // lecture octet divis en 2 valeurs hexa (poids fort et faible)
      b:=ReadByte;                 
      o:=b and $10;                // type of Table : 0=DC 1=AC
      b:=b and $0F;                // Huffman Table Number
      // HTable pointera sur la table  charger
      if o>0
      then HTable:=@FAcTable[b]
      else HTable:=@FDcTable[b];
      // lecture de la table
      // d'abord le nombre de codes de chaque taille
      j:=0;
      for i:=0 to 15
      do begin
        b:=ReadByte;
        HTable^.Bits[i]:=b;        // nb de codes de i bits
        Inc(j,b);                  // j = nb total de codes
      end;
      Dec(SectionLengthLeft, 17);  // on vient de lire 17 octets
      // puis les valeurs codes, dans l'ordre de l'arbre
      i:=0;
      while i<j
      do begin
        HTable^.Hval[i]:=ReadByte; // valeur reprsente par le (i+1)me code
        Inc(i);
      end;
      Dec(SectionLengthLeft, j);   // on vient de lire j octets
      // reconstruction du codage (les codes ne sont pas donns)
      k:=0;
      w:=0;                    // initialisation du code
      for i:=1 to 16           // on balaie les tailles de codes
      do begin
        j:=0;
        while j<HTable^.Bits[i-1] // tant qu'il en reste de cette catgorie (=taille)
        do begin                  // on les attribue
          HTable^.Size[k]:=i;  // taille en bits du (k+1)me code
          HTable^.Code[k]:=w;  // (k+1)me code
          Inc(w);              // code suivant
          Inc(k);
          Inc(j);
        end;
        w:=w shl 1; // le code suivant n'est pas attribu => sert de racine  ceux plus longs
      end;
    end;
  end;
// dcodage de la section SOS = tables de Huffman utilises par les diffrentes composantes
procedure Decode_SOS;
  var
    b,o,i: Byte;
  begin
    if (FComponentsNumber<>ReadByte) // 3 composantes dans le Scan => entrelac => OK
    then raise EdyJpegError.Create('Non Interleaved Components Scan Error');
    if (SectionLengthLeft<>2*FComponentsNumber+4)  // erreur de longueur
    then raise EdyJpegError.Create('Components Length Scan Error');
    // lecture des index des tables de Huffman par canal
    for i:=ccY to ccV
    do begin
      b:=ReadByte; // Component Id
      o:=ReadByte; // AC and DC Tables
      with FColorChannels[i]
      do begin
        DC_Table:=o shr 4;
        AC_Table:=o and $0F;
        if (Id<>b) or(DC_Table>2) or (AC_Table>2)
        then raise EdyJpegError.Create('Components ACDC Tables Error');
      end;
    end;
    // lecture donnes Progressive Jpeg
    with FProgressiveScan
    do begin
      ss:=ReadByte;
      se:=ReadByte;
      b:=ReadByte;
      ah:=b shr 4;
      al:=b and $0F;
    end;
    Terminated:=True; // le flux commence immdiatement aprs le SOS
  end;
// dcodage de la section DQT = table de quantization
procedure Decode_DQT;
  var
    b,o,i: Byte;
  begin
    while SectionLengthLeft>0
    do begin
      b:=ReadByte;
      o:=b shr 4;   // Precision : 0=8 bits 1=16 bits
      b:=b and $0F; // Quantization Table Number
      // chargement de la table
      for i:=0 to 63
      do if o>0
      then FQTable[b,i]:=ReadWord
      else FQTable[b,i]:=ReadByte;
      if o>0
      then Dec(SectionLengthLeft, 129)
      else Dec(SectionLengthLeft, 65);
    end;
  end;
// dcodage de la section DRI = dfinit si interruptions rgulires
// du flux ECS compress par des RSTn (marqueurs Restart)
// au bout de x MCU codes (dcale le cadre de lecture)
procedure Decode_DRI;
  begin
    FRstInterval:=ReadWord;
  end;
begin
  SectionLengthLeft:=0;
  Terminated:=False;
  while not Terminated
  do begin
    marker:=ReadByte;      // lecture du 1 octet du marqueur
    if marker<>$FF         // ce n'est pas un marqueur !
    then raise EdyJpegError.Create('Invalid JPEG File : marker wanted at '+IntToStr(FStream.Position-1))
    else marker:=ReadByte; // lecture du 2 octet du marqueur
    case marker of
      $D8 : Continue;      // Start of Image => lire le prochain marqueur
      $D9 : begin          // End of Image inattendu
              S:='Invalid JPEG File : EOI encountered at '+IntToStr(FStream.Position-1);
              raise EdyJpegError.Create(S);
            end;
    else SectionLengthLeft:=ReadWord-2; // lecture de la longueur du reste de la section
    end;
    // traitement de la section
    case marker of
       $C0 : Decode_SOF0;  // Start Of Frame 0 (composantes couleurs des Baseline Jpeg)
       $C1 : raise EdyJpegError.Create('Extended sequential Huffman coding, not supported');
       $C2 : raise EdyJpegError.Create('Progressive Huffman coding, not supported');
       $C3 : raise EdyJpegError.Create('Lossless Huffman coding, not supported');
       $C4 : Decode_DHT;   // Define Huffman Table (tables de dcompression)
       $C5 : raise EdyJpegError.Create('Differential sequential Huffman coding, not supported');
       $C6 : raise EdyJpegError.Create('Differential progressive Huffman coding, not supported');
       $C7 : raise EdyJpegError.Create('Differential lossless Huffman coding, not supported');
       $C8 : raise EdyJpegError.Create('Reserved for JPEG extensions, not supported');
       $C9 : raise EdyJpegError.Create('Extended sequential arithmetic coding, not supported');
       $CA : raise EdyJpegError.Create('Progressive arithmetic coding, not supported');
       $CB : raise EdyJpegError.Create('Lossless arithmetic coding, not supported');
       $CD : raise EdyJpegError.Create('Differential sequential arithmetic coding, not supported');
       $CE : raise EdyJpegError.Create('Differential progressive arithmetic coding, not supported');
       $CF : raise EdyJpegError.Create('Differential lossless arithmetic, not supported');
       $DA : Decode_SOS;   // Start Of Scan (attribue les tables de Huffman aux canaux) : prcde les donnes
       $DB : Decode_DQT;   // Define Quantization Table (limite les dtails)
       $DD : Decode_DRI;   // Define Restart Interval (interruptions rgulires du flux)       
    else FStream.Seek(SectionLengthLeft,soFromCurrent); // section inconnue => on la saute !
    end;
  end;
end;

procedure TdyYCbCrBaselineJPEGDecoder.DecodeFromStream(AStream: TStream);
var
  ComponentIndex: Byte;
  ScaleH, ScaleV: array[ccY..ccV] of Byte;
  CurrentDcTable, CurrentAcTable: THuffmanTable;
  CurrentQtTable: Byte;
  CurrentDcValue,_Y,_U,_V: SmallInt;
  Vector: array[0..63] of SmallInt;
  AcTerm: TAcTerm;
  DCT: TDctArray;
  Value,R,G,B: TIntToByte;
  PBA: PByteArray;
  PRT: TPRT absolute PBA; // pour traiter le buffer par octet, RGBTriple, ou YUV
  RestartMCU: Word;
  o,index,i,j,k,l,v,h,VScale,HScale: Byte;
  w,x,y,xx,yy: Word;
  c: Cardinal;
  ScanlineSize: integer;
begin
  FTicks:=GetTickCount; // initialisation chronomtre
  FStream:=AStream;
  // chargement des donnes et tables utiles au dcodage
  ReadMarkers;
  // SubSampling pour chaque canal et direction,
  // cad nombre de pixels cods par pixel stock,
  // partant du principe que Y est le plus prcis,
  // l'oeil tant plus sensible  la luminance qu'
  // la couleur (bcp plus de btonnets que de cnes)
  for i:=ccY to ccV
  do begin
    ScaleH[i]:=FYUV[ccY].HRes div FYUV[i].HRes;
    ScaleV[i]:=FYUV[ccY].VRes div FYUV[i].VRes;
  end;
  if Assigned(FBitmap)
  then FreeAndNil(FBitmap);
  FBitmap:=Graphics.TBitmap.Create;
  with FBitmap
  do begin
    PixelFormat:=pf24Bit;
    Width:=FWidth;
    Height:=FHeight;
    ScanlineSize:=Cardinal(ScanLine[1])-Cardinal(ScanLine[0]); // > 0 si Bottom-Up
    w:=FWidth*BytesPerPixel;
    if w>Abs(ScanlineSize) // vrification pour la copie des lignes
    then raise EdyJpegError.Create('Bitmap Scanline Width Error')
    else ScanlineSize:=w;
  end;
  // espace image
  GetMem(PBA,FHeight*ScanlineSize);
  FBuffer:=PBA; // pour qu'il soit libr au Clear (et donc au Destroy)
  // dcompression par MCU
  RestartMCU:=0;
  y:=0;
  while y<FHeight
  do begin
    x:=0;
    while x<FWidth
    do begin
      // et par composante : on balaye les canaux Y -> U -> V
      for ComponentIndex:=ccY to ccV
      do begin
        // initialisations diverses
        VScale:=ScaleV[ComponentIndex];
        HScale:=ScaleH[ComponentIndex];
        CurrentDcTable:=FDcTable[FColorChannels[ComponentIndex].DC_Table];
        CurrentAcTable:=FAcTable[FColorChannels[ComponentIndex].AC_Table];
        CurrentQtTable:=FColorChannels[ComponentIndex].QT_Table;
        CurrentDcValue:=FColorChannels[ComponentIndex].DcValue;
        // balayage des Data Units pour cette composante
        for v:=0 to FYUV[ComponentIndex].VRes-1 do
          for h:=0 to FYUV[ComponentIndex].HRes-1
          // dcodage d'une Data Unit
          do begin
            FillChar(Vector, SizeOf(Vector), 0);
            // lecture du coefficient DC (Direct : entropie du bloc, sorte de moyenne des pixels)
            o:=GetDcTermCategory(CurrentDcTable); // longueur du code
            w:=ReadBitsToWord(o);                 // code lui-mme
            Inc(CurrentDcValue, BitsToInt(o, w)); // traduction et ajout du diffrentiel de DC
            // en effet, ses variations sont codes : Differential Pulse Code Modulation
            Vector[0]:=CurrentDcValue;
            // lecture des termes AC (Alternative : relatifs aux frquences spatiales, cad aux dtails)
            index:=1;
            while index<64
            do begin
              AcTerm:=GetAcTerm(CurrentAcTable);
              if AcTerm.EOB=0
              then Break // fin de bloc, code par $00 => plus que des 0, or Vector en est dj rempli
              else begin // dcodage du coefficient AC
                Inc(index, AcTerm.Zeros);       // on saute les coefficients nuls (Zero Run Length Coding)
                o:=AcTerm.Categ;                // longueur du code
                w:=ReadBitsToWord(o);           // code lui-mme
                Vector[index]:=BitsToInt(o,w);  // traduction du coefficient
                Inc(index);
              end;
            end;
            // dequantization par multiplication avec la table
            // (la quantization limine des dtails par division)
            // et pr-calcul par produit avec un facteur pour l'IDCT
            for index:=0 to 63
            do begin
              k:=ZigZag[index];
              DCT[index]:=Vector[k]*FQTable[CurrentQtTable,k]*AAN_PreCalc[index];
            end;  // for index
            // Inverse Discrete Cosine Transformation
            // qui transforme les frquences spatiales en valeurs
            // sorte de transforme de Fourier inverse  2 dimensions
            AAN_IDCT(DCT);
            // dnormalisation des valeurs obtenues
            for index:=0 to 63
            do begin
              // dnormalisation des valeurs obtenues
              Value.Int:=Trunc(DCT[index]/8)+128; // /8 = facteur post-IDCT et +128 = Level Shift
              if Value.Int<0
              then Value.Int:=0
              else if Value.Int>255
              then Value.Int:=255;
              // affectation de la valeur au canal, pour tous les pixels reprsents,
              // ce qui dpend du niveau de SubSampling pour la composante considre
              for l:=0 to VScale-1 do
                for k:=0 to HScale-1
                do begin
                  xx := x + ( 8 * h + index mod 8 ) * HScale + k;
                  yy := y + ( 8 * v + index div 8 ) * VScale + l;
                  if (xx<FWidth) and (yy<FHeight) // accs comme un tableau d'octets, grce  PBA
                  then PBA[ (yy * FWidth + xx) * BytesPerPixel + ComponentIndex-1 ] := Value.Byte;
                end; // for k,l
            end; // for index
          end; // for v,h
        FColorChannels[ComponentIndex].DcValue:=CurrentDcValue; // base pour le DC de la prochaine DU
      end; // for ComponentIndex
      if FRstInterval>0  // dans ce cas, on prend en compte le nombre de MCU dcodes
      then begin         // pour sauter le marqueur Restart (RSTn) attendu au prochain mot
        Inc(RestartMCU);
        if RestartMCU=FRstInterval
        then begin
          RestartMCU:=0;      // on repart pour un nouvel intervalle
          FNextBitToRead:=0;  // pour sauter la fin de l'octet en cours (le RSTn sera saut lors de la lecture du suivant)
          // chaque intervalle redmarre avec un codage DC complet pour ne pas dpendre du prcdent
          for ComponentIndex:=ccY to ccV do FColorChannels[ComponentIndex].DcValue:=0;
        end;
      end;
      Inc(x,FMCU.Width);
    end; // while x
    Inc(y,FMCU.Height);
  end; // while y
  if ReadWord<>$FFD9 // le fichier doit se terminer par un marqueur EOI : End Of Image (interrompt les RSTn)
  then raise EdyJpegError.Create('Invalid JPEG File : EOI marker wanted at '+IntToStr(FStream.Position-2));
  // traduction YUV -> RGB et passage au Bitmap
  for yy:=0 to FHeight-1 do
    for xx:=0 to FWidth-1
    do begin
      c:=yy*FWidth+xx;
      with PRT[c], PRT[c].Triple // accs comme un tableau d'YUV ou de RGB grce  PRT
      do begin
        // pr_calculs
        _Y:=Y;
        _U:=U-128;
        _V:=V-128;
        // rouge
        R.Int := Trunc(_Y + 1.40200 * _V);
        if R.Int<0
        then R.Int:=0
        else if R.Int>255
        then R.Int:=255;
        rgbtRed  := R.Byte;
        // vert
        G.Int := Trunc(_Y - 0.34414 * _U - 0.71414 * _V);
        if G.Int<0
        then G.Int:=0
        else if G.Int>255
        then G.Int:=255;
        rgbtGreen := G.Byte;
        // bleu
        B.Int := Trunc(_Y + 1.77200 * _U);
        if B.Int<0
        then B.Int:=0
        else if B.Int>255
        then B.Int:=255;
        rgbtBlue  := B.Byte;
      end;
    end;
  // passage au Bitmap
  for yy:=0 to FHeight-1
  do Move( PRT[yy*FWidth], TPRT(FBitmap.ScanLine[yy])[0], ScanlineSize );
  FTicks:=GetTickCount-FTicks; // temps pass  dcompresser
end;

initialization
  // initialisation des constantes pour l'IDCT
  AAN[ct1]:=1.414213562;
  AAN[ct2]:=1.847759065;
  AAN[ct3]:=1.082392200;
  AAN[ct4]:=-2.613125930;
  // pr-calcul des facteurs aprs DQT et avant IDCT
  for k:=0 to 63
  do begin
    i:=k mod 8;
    j:=k div 8;
    AAN_PreCalc[k]:=AAN_Scale[j]*AAN_Scale[i];
  end;

finalization

end.
